""" tests for fnctionality in super class"""
import copy
import os
import pickle

import joblib
import numpy as np
from sklearn import datasets

from aisdc.safemodel.reporting import get_reporting_string
from aisdc.safemodel.safemodel import SafeModel

notok_start = get_reporting_string(name="warn_possible_disclosure_risk")
ok_start = get_reporting_string(name="within_recommended_ranges")


class DummyClassifier:
    """Dummy Classifier that always returns predictions of zero"""

    # pylint: disable=too-many-arguments
    def __init__(
        self, at_least_5f=5.0, at_most_5i=5, exactly_boo="boo", keyA=True, keyB=True
    ):
        """dummy init"""
        self.at_least_5f = at_least_5f
        self.at_most_5i = at_most_5i
        self.exactly_boo = exactly_boo
        self.keyA = keyA
        self.keyB = keyB

    def fit(self, x: np.ndarray, y: np.ndarray):
        """dummy fit"""

    def predict(self, x: np.ndarray):  # pragma: no cover
        """predict all ones"""
        return np.ones(x.shape[0])


def get_data():
    """Returns data for testing."""
    iris = datasets.load_iris()
    x = np.asarray(iris["data"], dtype=np.float64)
    y = np.asarray(iris["target"], dtype=np.float64)
    x = np.vstack([x, (7, 2.0, 4.5, 1)])
    y = np.append(y, 4)
    return x, y


class SafeDummyClassifier(
    SafeModel, DummyClassifier
):  # pylint:disable=too-many-instance-attributes
    """Privacy protected dummy classifier."""

    def __init__(self, **kwargs) -> None:
        """Creates model and applies constraints to params."""
        SafeModel.__init__(self)
        self.basemodel_paramnames = (
            "at_least_5f",
            "at_most_5i",
            "exactly_boo",
            "keyA",
            "keyB",
        )
        the_kwds = {}
        for key, val in kwargs.items():
            if key in self.basemodel_paramnames:
                the_kwds[key] = val
        DummyClassifier.__init__(self, **the_kwds)
        self.model_type: str = "DummyClassifier"
        self.ignore_items = ["model_save_file", "basemodel_paramnames", "ignore_items"]
        # create an item to test additional_checks()
        self.examine_seperately_items = ["newthing"]
        self.newthing = ["myStringKey", "aString", "myIntKey", "42"]

    def set_params(self, **kwargs):  # pragma: no cover
        """sets params"""
        for key, val in kwargs.items():  # pylint:disable=unused-variable
            self.key = val  # pylint:disable=attribute-defined-outside-init

    def fit(self, x: np.ndarray, y: np.ndarray):
        """dummy fit"""
        self.saved_model = copy.deepcopy(self.__dict__)


def test_params_checks_ok():
    """test parameter  checks ok"""
    model = SafeDummyClassifier()

    correct_msg = ok_start
    msg, disclosive = model.preliminary_check()
    print(
        f"exactly_boo is {model.exactly_boo} with type{type(model.exactly_boo).__name__}"
    )
    assert msg == ok_start, f"Correct msg:\n{correct_msg}\nActual msg:\n{msg}\n"
    assert disclosive is False


def test_params_checks_too_low():
    """test parameter  checks too low"""
    model = SafeDummyClassifier()

    model.at_least_5f = 4.0
    msg, disclosive = model.preliminary_check()
    assert disclosive is True
    correct_msg = notok_start + get_reporting_string(
        name="less_than_min_value",
        key="at_least_5f",
        cur_val=model.at_least_5f,
        val=5.0,
    )
    assert msg == correct_msg, f"Correct msg:\n{correct_msg}\nActual msg:\n{msg}\n"


def test_params_checks_too_high():
    """test parameter  checks too high"""
    model = SafeDummyClassifier()

    model.at_most_5i = 6
    msg, disclosive = model.preliminary_check()
    assert disclosive is True
    correct_msg = notok_start + get_reporting_string(
        name="greater_than_max_value", key="at_most_5i", cur_val=model.at_most_5i, val=5
    )
    assert msg == correct_msg, f"Correct msg:\n{correct_msg}\nActual msg:\n{msg}\n"


def test_params_checks_not_equal():
    """test parameter  checks not equal"""
    model = SafeDummyClassifier()

    model.exactly_boo = "foo"
    msg, disclosive = model.preliminary_check()
    assert disclosive is True
    correct_msg = notok_start + get_reporting_string(
        name="different_than_fixed_value",
        key="exactly_boo",
        cur_val=model.exactly_boo,
        val="boo",
    )
    assert msg == correct_msg, f"Correct msg:\n{correct_msg}\nActual msg:\n{msg}\n"


def test_params_checks_wrong_type_str():
    """test parameter  checks wrong type - strings given"""
    model = SafeDummyClassifier()

    model.at_least_5f = "five"
    model.at_most_5i = "five"

    msg, disclosive = model.preliminary_check()
    assert disclosive is True
    correct_msg = notok_start
    correct_msg += get_reporting_string(
        name="different_than_recommended_type",
        key="at_least_5f",
        cur_val=model.at_least_5f,
        val=5.0,
    )
    correct_msg += get_reporting_string(
        name="different_than_recommended_type",
        key="at_least_5f",
        cur_val=model.at_least_5f,
        val="float",
    )
    correct_msg += get_reporting_string(
        name="different_than_recommended_type",
        key="at_most_5i",
        cur_val=model.at_most_5i,
        val=5,
    )
    correct_msg += get_reporting_string(
        name="different_than_recommended_type",
        key="at_most_5i",
        cur_val=model.at_most_5i,
        val="int",
    )

    assert msg == correct_msg, f"Correct msg:\n{correct_msg}\nActual msg:\n{msg}\n"


def test_params_checks_wrong_type_float():
    """test parameter  checks wrong_type_float"""
    model = SafeDummyClassifier()

    model.exactly_boo = 5.0
    model.at_most_5i = 5.0

    _, disclosive = model.preliminary_check()
    assert disclosive is True
    correct_msg = notok_start

    correct_msg += get_reporting_string(
        name="different_than_recommended_type",
        key="at_most_5",
        cur_val=model.at_most_5i,
        val="int",
    )
    correct_msg += get_reporting_string(
        name="different_than_recommended_type",
        key="exactly_5",
        cur_val=model.exactly_boo,
        val="int",
    )


def test_params_checks_wrong_type_int():
    """test parameter  checks wrong_type_intt"""
    model = SafeDummyClassifier()

    model.exactly_boo = 5
    model.at_least_5f = 5

    msg, disclosive = model.preliminary_check()
    assert disclosive is True
    correct_msg = notok_start

    correct_msg += get_reporting_string(
        name="different_than_recommended_type",
        key="at_least_5f",
        cur_val=model.at_least_5f,
        val="float",
    )
    correct_msg += get_reporting_string(
        name="different_than_recommended_type",
        key="exactly_boo",
        cur_val=model.exactly_boo,
        val="str",
    )
    correct_msg += get_reporting_string(
        name="different_than_fixed_value",
        key="exactly_boo",
        cur_val=model.exactly_boo,
        val="boo",
    )
    assert msg == correct_msg, f"Correct msg:\n{correct_msg}\nActual msg:\n{msg}\n"


def test_check_unknown_param():
    """checks handling of malformed json rule"""
    # pylint:disable=protected-access,no-member
    model = SafeDummyClassifier()
    _, _ = model.preliminary_check()
    odd_rule = {"operator": "unknown", "keyword": "exactly_boo", "value": "some_val"}
    msg, disclosive = model._SafeModel__check_model_param(odd_rule, False)
    correct_msg = get_reporting_string(
        name="unknown_operator",
        key=odd_rule["keyword"],
        val=odd_rule["value"],
        cur_val="boo",
    )
    assert msg == correct_msg, f"Correct msg:\n{correct_msg}\nActual msg:\n{msg}\n"
    assert disclosive is False


def test_check_model_param_or():
    """tests or conditions in rules.json
    the and condition is tested by the decision tree tests
    """
    # ok
    model = SafeDummyClassifier()
    msg, disclosive = model.preliminary_check()
    correct_msg = ok_start
    assert msg == correct_msg, f"Correct msg:\n{correct_msg}\nActual msg:\n{msg}\n"
    assert disclosive is False

    part1 = get_reporting_string(
        name="different_than_fixed_value",
        key="keyA",
        cur_val=False,
        val="True",
    )
    part2 = get_reporting_string(
        name="different_than_fixed_value",
        key="keyB",
        cur_val=False,
        val="True",
    )

    # or - branch 1
    model = SafeDummyClassifier(keyA=False)
    correct_msg = ok_start + part1
    msg, disclosive = model.preliminary_check()
    assert msg == correct_msg, f"Correct msg:\n{correct_msg}\nActual msg:\n{msg}\n"
    assert disclosive is False

    # or  branch 2
    model = SafeDummyClassifier(keyB=False)
    correct_msg = ok_start + part2
    msg, disclosive = model.preliminary_check()
    assert msg == correct_msg, f"Correct msg:\n{correct_msg}\nActual msg:\n{msg}\n"
    assert disclosive is False

    # fail or
    model = SafeDummyClassifier(keyA=False, keyB=False)
    correct_msg = notok_start + part1 + part2
    msg, disclosive = model.preliminary_check()
    assert msg == correct_msg, f"Correct msg:\n{correct_msg}\nActual msg:\n{msg}\n"
    assert disclosive is True


def test_saves():
    """checks that save functions as expected"""
    model = SafeDummyClassifier()
    x, y = get_data()
    model.fit(x, y)

    # no name provided
    model.save()

    # no extension
    model.save("mymodel")

    # unsupported extension
    model.save("mymodel.unsupported")

    # cannot be pickled
    model.square = lambda x: x * x  # pylint: disable=attribute-defined-outside-init
    model.save("unpicklable.pkl")

    # cannot be joblibbed
    model.square = lambda x: x * x  # pylint: disable=attribute-defined-outside-init
    model.save("unpicklable.sav")

    # cleanup
    for name in ("dummy.pkl", "dummy.sav", "unpicklable.pkl", "unpicklable.sav"):
        if os.path.exists(name) and os.path.isfile(name):
            os.remove(name)


def test_loads():
    """basic check that making, changing,saving,loading model works"""
    model = SafeDummyClassifier()
    x, y = get_data()
    model.fit(x, y)
    # change something in model
    model.exactly_boo = "this_should_be_present"
    assert model.exactly_boo == "this_should_be_present"

    # pkl
    model.save("dummy.pkl")
    with open("dummy.pkl", "rb") as file:
        model2 = pickle.load(file)
    assert model2.exactly_boo == "this_should_be_present"

    # joblib
    model.save("dummy.sav")
    with open("dummy.sav", "rb") as file:
        model2 = joblib.load(file)
    assert model2.exactly_boo == "this_should_be_present"

    # cleanup
    for name in ("dummy.pkl", "dummy.sav"):
        if os.path.exists(name) and os.path.isfile(name):
            os.remove(name)


def test__apply_constraints():
    """tests constraints can be applied as expected"""

    # wrong type
    model = SafeDummyClassifier()
    model.at_least_5f = 3.9
    model.at_most_5i = 6.2
    model.exactly_boo = "five"

    assert model.at_least_5f == 3.9
    assert model.at_most_5i == 6.2
    assert model.exactly_boo == "five"

    msg, _ = model.preliminary_check(verbose=True, apply_constraints=True)

    assert model.at_least_5f == 5.0
    assert model.at_most_5i == 5
    assert model.exactly_boo == "boo"

    # checks that type changes happen correctly
    model = SafeDummyClassifier()
    model.at_least_5f = int(6.0)
    model.at_most_5i = float(4.2)
    model.exactly_boo = "five"

    assert model.at_least_5f == int(6)
    assert model.at_most_5i == 4.2
    assert model.exactly_boo == "five"

    msg, _ = model.preliminary_check(verbose=True, apply_constraints=True)
    assert model.at_least_5f == 6.0
    assert model.at_most_5i == 4
    assert model.exactly_boo == "boo"

    correct_msg = notok_start
    correct_msg += get_reporting_string(
        name="different_than_recommended_type",
        key="at_least_5f",
        cur_val=int(6),
        val="float",
    )
    correct_msg += get_reporting_string(
        name="change_param_type", key="at_least_5f", val="float"
    )
    correct_msg += get_reporting_string(
        name="different_than_recommended_type",
        key="at_most_5i",
        cur_val=4.2,
        val="int",
    )
    correct_msg += get_reporting_string(
        name="change_param_type", key="at_most_5i", val="int"
    )
    correct_msg += get_reporting_string(
        name="different_than_fixed_value",
        key="exactly_boo",
        cur_val="five",
        val="boo",
    )
    correct_msg += get_reporting_string(
        name="changed_param_equal", key="exactly_boo", val="boo"
    )

    assert msg == correct_msg, f"Correct msg:\n{correct_msg}\nActual msg:\n{msg}\n"

    model = SafeDummyClassifier()
    adict = {"a": "67543", "b": 6542}
    model.exactly_boo = adict
    msg, _ = model.preliminary_check(verbose=True, apply_constraints=True)
    correct_msg = notok_start
    correct_msg += get_reporting_string(
        name="different_than_recommended_type",
        key="exactly_boo",
        cur_val={"a": "67543", "b": 6542},
        val="str",
    )
    correct_msg += get_reporting_string(
        name="not_implemented_for_change", key="exactly_boo", cur_val=adict, val="str"
    )
    correct_msg += get_reporting_string(
        name="different_than_fixed_value",
        key="exactly_boo",
        cur_val={"a": "67543", "b": 6542},
        val="boo",
    )
    correct_msg += get_reporting_string(
        name="changed_param_equal", key="exactly_boo", val="boo"
    )
    assert msg == correct_msg, f"Correct msg:\n{correct_msg}\nActual msg:\n{msg}\n"


def test_get_saved_model_exception():
    """tests the exception handling
    in get_current_and_saved_models()
    """
    model = SafeDummyClassifier()
    # add generator which can't be pickled or copied

    model.a_generator = (  # pylint: disable=attribute-defined-outside-init
        i for i in [1, 2, 3]
    )
    current, saved = model.get_current_and_saved_models()
    assert saved == {}  # since we haven;t called fit()
    assert (  # pylint: disable=consider-iterating-dictionary
        "a_generator" not in current.keys()
    )


def test_generic_additional_tests():
    """checks the class generic additional tests
    for this purpose SafeDummyClassifier()
    defines
    self.newthing = {"myStringKey": "aString", "myIntKey": 42}
    """
    model = SafeDummyClassifier()
    x, y = get_data()
    model.fit(x, y)

    # ok
    msg, disclosive = model.posthoc_check()
    correct_msg = ""
    assert msg == correct_msg, f"Correct msg:\n{correct_msg}\nActual msg:\n{msg}\n"
    assert disclosive is False

    # different lengths
    model.saved_model["newthing"] += ("extraA",)
    msg, disclosive = model.posthoc_check()
    print(
        "contents of new then saved\n"
        f"{model.newthing}\n"
        f'{model.saved_model["newthing"]}'
    )

    correct_msg = "Warning: different counts of values for parameter newthing.\n"
    assert msg == correct_msg, f"Correct msg:\n{correct_msg}\nActual msg:\n{msg}\n"
    assert disclosive is True

    # different thing in list
    model.newthing += ("extraB",)
    msg, disclosive = model.posthoc_check()
    correct_msg = (
        "Warning: at least one non-matching value for parameter list newthing.\n"
    )
    print(
        "contents of new then saved\n"
        f"{model.newthing}\n"
        f'{model.saved_model["newthing"]}'
    )
    assert msg == correct_msg, f"Correct msg:\n{correct_msg}\nActual msg:\n{msg}\n"
    assert disclosive is True


def test_request_release():
    """checks requestrelease code works"""
    model = SafeDummyClassifier()
    x, y = get_data()
    model.fit(x, y)
    # give it k_anonymity
    model.k_anonymity = 5  # pylint: disable=attribute-defined-outside-init

    # no file provided, has k_anonymity
    model.request_release()



"""This module contains unit tests for the SafeRandomForestClassifier."""

import copy
import os
import pickle

import joblib
import numpy as np
from sklearn import datasets

from aisdc.safemodel.classifiers import SafeRandomForestClassifier
from aisdc.safemodel.reporting import get_reporting_string

EXPECTED_ACC = 0.9470198675496688  # 5 estimators, min_samples_leaf 5


class DummyClassifier:
    """Dummy Classifier that always returns predictions of zero"""

    def __init__(self):
        """empty init"""

    def fit(self, x: np.ndarray, y: np.ndarray):
        """empty fit"""

    def predict(self, x: np.ndarray):
        """predict all ones"""

    #     return np.ones(x.shape[0])


def get_data():
    """Returns data for testing."""
    iris = datasets.load_iris()
    x = np.asarray(iris["data"], dtype=np.float64)
    y = np.asarray(iris["target"], dtype=np.float64)
    x = np.vstack([x, (7, 2.0, 4.5, 1)])
    y = np.append(y, 4)
    return x, y


def test_randomforest_unchanged():
    """SafeRandomForestClassifier using recommended values."""
    x, y = get_data()
    model = SafeRandomForestClassifier(random_state=1, n_estimators=5)
    model.fit(x, y)
    assert model.score(x, y) == EXPECTED_ACC
    msg, disclosive = model.preliminary_check()
    correct_msg = get_reporting_string(name="within_recommended_ranges")
    assert msg == correct_msg, f"{msg}\n should be {correct_msg}"
    assert disclosive is False


def test_randomforest_recommended():
    """SafeRandomForestClassifier using recommended values."""
    x, y = get_data()
    model = SafeRandomForestClassifier(random_state=1, n_estimators=5)
    model.min_samples_leaf = 6
    model.fit(x, y)
    msg, disclosive = model.preliminary_check()
    correct_msg = "Model parameters are within recommended ranges.\n"
    assert msg == correct_msg, f"{msg}\n should be {correct_msg}"
    assert disclosive is False


def test_randomforest_unsafe_1():
    """SafeRandomForestClassifier with unsafe changes."""
    x, y = get_data()
    model = SafeRandomForestClassifier(random_state=1, n_estimators=5)
    model.bootstrap = False
    model.fit(x, y)
    msg, disclosive = model.preliminary_check()
    correct_msg = (
        "WARNING: model parameters may present a disclosure risk:\n"
        "- parameter bootstrap = False identified as different than the recommended "
        "fixed value of True."
    )
    assert msg == correct_msg, f"{msg}\n should be {correct_msg}"
    assert disclosive is True


def test_randomforest_unsafe_2():
    """SafeRandomForestClassifier with unsafe changes."""
    model = SafeRandomForestClassifier(random_state=1, n_estimators=5)
    model.bootstrap = True
    model.min_samples_leaf = 2
    msg, disclosive = model.preliminary_check()
    correct_msg = (
        "WARNING: model parameters may present a disclosure risk:\n"
        "- parameter min_samples_leaf = 2 identified as less than the recommended "
        "min value of 5."
    )
    assert msg == correct_msg, f"{msg}\n should be {correct_msg}"
    assert disclosive is True


def test_randomforest_unsafe_3():
    """SafeRandomForestClassifier with unsafe changes."""
    model = SafeRandomForestClassifier(random_state=1, n_estimators=5)
    model.bootstrap = False
    model.min_samples_leaf = 2
    msg, disclosive = model.preliminary_check()
    correct_msg = (
        "WARNING: model parameters may present a disclosure risk:\n"
        "- parameter bootstrap = False identified as different than the recommended "
        "fixed value of True."
        "- parameter min_samples_leaf = 2 identified as less than the recommended "
        "min value of 5."
    )
    assert msg == correct_msg, f"{msg}\n should be {correct_msg}"
    assert disclosive is True


def test_randomforest_save():
    """SafeRandomForestClassifier model saving."""
    x, y = get_data()
    model = SafeRandomForestClassifier(random_state=1, n_estimators=5)
    model.fit(x, y)
    # test pickle
    model.save("rf_test.pkl")
    with open("rf_test.pkl", "rb") as file:
        pkl_model = pickle.load(file)
    assert pkl_model.score(x, y) == EXPECTED_ACC
    # test joblib
    model.save("rf_test.sav")
    with open("rf_test.sav", "rb") as file:
        sav_model = joblib.load(file)
    assert sav_model.score(x, y) == EXPECTED_ACC

    # cleanup
    for name in ("rf_test.pkl", "rf_test.sav"):
        if os.path.exists(name) and os.path.isfile(name):
            os.remove(name)


def test_randomforest_hacked_postfit():
    """SafeRandomForestClassifier changes made to parameters after fit() called."""
    x, y = get_data()
    model = SafeRandomForestClassifier(random_state=1, n_estimators=5)
    model.bootstrap = False
    model.fit(x, y)
    model.bootstrap = True
    # preliminary check looks ok
    msg, disclosive = model.preliminary_check()
    correct_msg = get_reporting_string(name="within_recommended_ranges")
    assert msg == correct_msg, f"{msg}\n should be {correct_msg}"
    assert disclosive is False
    # but more detailed analysis says not
    msg2, disclosive2 = model.posthoc_check()
    part1 = get_reporting_string(name="basic_params_differ", length=1)
    part2 = get_reporting_string(
        name="param_changed_from_to", key="bootstrap", val=False, cur_val=True
    )
    part3 = get_reporting_string(
        name="param_changed_from_to",
        key="base_estimator",
        val="DecisionTreeClassifier()",
        cur_val="DecisionTreeClassifier()",
    )
    correct_msg2 = part1 + part2 + part3
    # print(f'Correct: {correct_msg2}\n Actual: {msg2}')

    assert msg2 == correct_msg2, f"{msg2}\n should be {correct_msg2}"
    assert disclosive2 is True


def test_not_fitted():
    """posthoc_check() called on unfitred model
    could have anything injected in classifier parameters"""
    unfitted_model = SafeRandomForestClassifier(random_state=1, n_estimators=5)

    # not fitted
    msg, disclosive = unfitted_model.posthoc_check()
    part1 = get_reporting_string(name="error_not_called_fit")
    part2 = get_reporting_string(name="recommend_do_not_release")
    correct_msg = part1 + part2
    assert msg == correct_msg, f"{msg}\n should be {correct_msg}"
    assert disclosive


def test_randomforest_modeltype_changed():
    """model type has been changed after fit()
    in this this case to hide some data
    """
    x, y = get_data()
    model = SafeRandomForestClassifier(random_state=1, n_estimators=5)
    correct_msg = ""
    # check code that tests base_estimator_
    model.fit(x, y)
    model.base_estimator = "DummyClassifier()"

    # hide some data
    for i in range(5):  # changed lengths get picked up in different test
        model.estimators_[i] = x[i, :]

    msg, disclosive = model.posthoc_check()
    # correct_msg += get_reporting_string(name="basic_params_differ",length=1)
    correct_msg = get_reporting_string(name="forest_estimators_differ", idx=5)
    correct_msg += get_reporting_string(
        name="param_changed_from_to",
        key="base_estimator",
        val="DecisionTreeClassifier()",
        cur_val="DummyClassifier()",
    )
    #    correct_msg += ("structure base_estimator has 1 differences: [('change', '', "
    #                    "(DecisionTreeClassifier(), DecisionTreeClassifier()))]"
    #                   )
    print(f"Correct: {correct_msg} Actual: {msg}")

    assert msg == correct_msg, f"{msg}\n should be {correct_msg}"
    assert disclosive is True, "should have been flagged as disclosive"


def test_randomforest_hacked_postfit_trees_removed():
    """tests various combinations of removing trees"""
    x, y = get_data()
    model = SafeRandomForestClassifier(random_state=1, n_estimators=5)
    # code that checks estimators_ : one other or both missing or different number or size
    model.fit(x, y)

    # tree removed from current
    the_estimators = model.__dict__.pop("estimators_")
    msg, disclosive = model.posthoc_check()
    correct_msg = get_reporting_string(name="current_item_removed", item="estimators_")
    # print(f'Correct: {correct_msg} Actual: {msg}')
    assert disclosive, "should be flagged as disclosive"
    assert msg == correct_msg, f"{msg}\n should be {correct_msg}"

    # trees removed from both
    _ = model.saved_model.pop("estimators_")
    msg, disclosive = model.posthoc_check()
    correct_msg = get_reporting_string(name="both_item_removed", item="estimators_")
    # print(f'Correct: {correct_msg} Actual: {msg}')
    assert disclosive, "should be flagged as disclosive"
    assert msg == correct_msg, f"{msg}\n should be {correct_msg}"

    # trees just removed from saved
    model.estimators_ = the_estimators
    msg, disclosive = model.posthoc_check()
    correct_msg = get_reporting_string(name="saved_item_removed", item="estimators_")
    # print(f'Correct: {correct_msg} Actual: {msg}')
    assert disclosive, "should be flagged as disclosive"
    assert msg == correct_msg, f"{msg}\n should be {correct_msg}"


def test_randomforest_hacked_postfit_trees_swapped():
    """trees swapped with those from a different random forest"""
    x, y = get_data()
    model = SafeRandomForestClassifier(random_state=1, n_estimators=5)
    diffsizemodel = SafeRandomForestClassifier(
        random_state=1, n_estimators=5, max_depth=2
    )

    # code that checks estimators_ : one other or both missing or different number or size
    model.fit(x, y)
    diffsizemodel.fit(x, y)

    # swap saved models
    the_saved_model = copy.deepcopy(model.saved_model)
    the_saved_diffmodel = copy.deepcopy(diffsizemodel.saved_model)
    model.saved_model = copy.deepcopy(the_saved_diffmodel)
    diffsizemodel.saved_model = copy.deepcopy(the_saved_model)
    model.posthoc_check()
    msg, disclosive = diffsizemodel.posthoc_check()
    part1 = get_reporting_string(name="basic_params_differ", length=1)
    part2 = get_reporting_string(
        name="param_changed_from_to", key="max_depth", val="None", cur_val="2"
    )
    part3 = get_reporting_string(name="forest_estimators_differ", idx=5)
    part4 = get_reporting_string(
        name="param_changed_from_to",
        key="base_estimator",
        val="DecisionTreeClassifier()",
        cur_val="DecisionTreeClassifier()",
    )
    correct_msg = part1 + part2 + part3 + part4
    # print(f'Correct:\n{correct_msg} Actual:\n{msg}')
    assert msg == correct_msg, f"{msg}\n should be {correct_msg}"
    assert disclosive, "should be flagged as disclosive"


def test_randomforest_hacked_postfit_moretrees():
    """trees added after fit"""
    x, y = get_data()
    model = SafeRandomForestClassifier(random_state=1, n_estimators=5)
    diffsizemodel = SafeRandomForestClassifier(random_state=1, n_estimators=10)
    model.fit(x, y)
    diffsizemodel.fit(x, y)

    # swap saved models
    the_saved_model = copy.deepcopy(model.saved_model)
    the_saved_diffmodel = copy.deepcopy(diffsizemodel.saved_model)
    model.saved_model = copy.deepcopy(the_saved_diffmodel)
    diffsizemodel.saved_model = copy.deepcopy(the_saved_model)
    msg, disclosive = diffsizemodel.posthoc_check()
    part1 = get_reporting_string(name="basic_params_differ", length=1)
    part2 = get_reporting_string(
        name="param_changed_from_to", key="n_estimators", val="5", cur_val="10"
    )
    part3 = get_reporting_string(name="different_num_estimators", num1=10, num2=5)
    part4 = get_reporting_string(
        name="param_changed_from_to",
        key="base_estimator",
        val="DecisionTreeClassifier()",
        cur_val="DecisionTreeClassifier()",
    )
    correct_msg = part1 + part2 + part3 + part4
    # print(f'Correct:\n{correct_msg} Actual:\n{msg}')
    assert msg == correct_msg, f"{msg}\n should be {correct_msg}"
    assert disclosive, "should be flagged as disclosive"



"""Code for automatic report generation"""
import abc
import json

import numpy as np
import pylab as plt
from fpdf import FPDF

# Adds a border to all pdf cells of set to 1 -- useful for debugging
BORDER = 0

DISPLAY_METRICS = [
    "AUC",
    "ACC",
    "Advantage",
    "FDIF01",
    "PDIF01",
    "TPR@0.1",
    "TPR@0.01",
    "TPR@0.001",
    "TPR@1e-05",
]

MAPPINGS = {"PDIF01": lambda x: np.exp(-x)}

INTRODUCTION = (
    "This report provides a summary of a series of simulated attack experiments performed "
    "on the model outputs provided. An attack model is trained to attempt to distinguish "
    "between outputs from training (in-sample) and testing (out-of-sample) data. The metrics "
    "below describe the success of this classifier. A successful classifier indicates that the "
    "original model is unsafe and should not be allowed to be released from the TRE.\n"
    "In particular, the simulation splits the data provided into test and train sets (each will "
    "in- and out-of-sample examples). The classifier is trained on the train set and evaluated "
    "on the test set. This is repeated with different train/test splits a user-specified number "
    "of times.\n"
    "To help place the results in context, the code may also have run a series of baseline "
    "experiments. In these, random model outputs for hypothetical in- and out-of-sample data are "
    "generated with identical statistical properties. In these baseline cases, there is no signal "
    "that an attacker could leverage and therefore these values provide a baseline against "
    "which the actual values can be compared.\n"
    "For some metrics (FDIF and AUC), we are able to compute p-values. In each case, shown below "
    "(in the Global metrics sections) is the number of repetitions that exceeded the p-value "
    "threshold both without, and with correction for multiple testing (Benjamini-Hochberg "
    "procedure).\n"
    "ROC curves for all real (red) and dummy (blue) repetitions are provided. These are shown in "
    "log space (as reommended here [ADD URL]) to emphasise the region in which risk is highest -- "
    "the bottom left (are high true positive rates possible with low false positive rates).\n"
    "A description of the metrics and how to interpret them within the context of an attack is "
    "given below."
)

LOGROC_CAPTION = (
    "This plot shows the False Positive Rate (x) versus the True Positive Rate (y). "
    "The axes are in log space enabling us to focus on areas where the False Positive Rate is low "
    "(left hand area). Curves above the y = x line (black dashes) in this region represent a "
    "disclosure risk as an attacker can obtain many more true than false positives. "
    "The solid coloured lines show the curves for the attack simulations with the true model "
    "outputs. The lighter grey lines show the curves for randomly generated outputs with "
    "no structure (i.e. in- and out-of- sample predictions are generated from the same "
    "distributions. Solid curves consistently higher than the grey curves in the left hand "
    "part of the plot are a sign of concern."
)

GLOSSARY = {
    "AUC": "Area Under the ROC curve",
    "True Positive Rate (TPR)": (
        "The true positive rate is the number of True Positives that are predicted as positive as "
        "a proportion of the total number of positives. If an attacker has N examples that were "
        "actually in the training set, the TPR is the proportion of these that they predict as "
        "being in the training set."
    ),
    "ACC": "The proportion of predictions that the attacker makes that are correct.",
}


class NumpyArrayEncoder(json.JSONEncoder):
    """Json encoder that can cope with numpy arrays"""

    def default(self, o):
        """If an object is an np.ndarray, convert to list"""
        if isinstance(o, np.ndarray):
            return o.tolist()
        if isinstance(o, np.int64):
            return int(o)
        if isinstance(o, np.int32):
            return int(o)
        if isinstance(o, abc.ABCMeta):
            return str(o)
        return json.JSONEncoder.default(self, o)


def _write_dict(pdf, input_dict, indent=0, border=BORDER):
    """Write a dictionary to the pdf"""
    for key, value in input_dict.items():
        pdf.set_font("arial", "B", 14)
        pdf.cell(75, 5, key, border, 1, "L")
        pdf.cell(indent, 0)
        pdf.set_font("arial", "", 12)
        pdf.multi_cell(150, 5, value, border, "L")
        pdf.ln(h=5)


def title(pdf, text, border=BORDER, font_size=24, font_style="B"):
    """Write a title block"""
    pdf.set_font("arial", font_style, font_size)
    pdf.ln(h=5)
    pdf.cell(0, 0, text, border, 1, "C")
    pdf.ln(h=5)


def subtitle(
    pdf, text, indent=10, border=BORDER, font_size=12, font_style="B"
):  # pylint: disable = too-many-arguments
    """Write a subtitle block"""
    pdf.cell(indent, border=border)
    pdf.set_font("arial", font_style, font_size)
    pdf.cell(75, 10, text, border, 1)


def line(
    pdf, text, indent=0, border=BORDER, font_size=11, font_style="", font="arial"
):  # pylint: disable = too-many-arguments
    """Write a standard block"""
    if indent > 0:
        pdf.cell(indent, border=border)
    pdf.set_font(font, font_style, font_size)
    pdf.multi_cell(0, 5, text, border, 1)


def _roc_plot_single(metrics, save_name):
    """Create a roc_plot for a single experiment"""
    plt.figure()
    plt.plot([0, 1], [0, 1], "k--")
    plt.plot(metrics["fpr"], metrics["tpr"], "r", linewidth=2)
    plt.xscale("log")
    plt.yscale("log")
    plt.grid()
    plt.xlabel("True Positive Rate")
    plt.ylabel("False Positive Rate")
    plt.tight_layout()
    plt.savefig(save_name)


def _roc_plot(metrics, dummy_metrics, save_name):
    """Create a roc plot for multiple repetitions"""
    plt.figure()
    plt.plot([0, 1], [0, 1], "k--")
    if dummy_metrics is None or len(dummy_metrics) == 0:
        do_dummy = False
    else:
        do_dummy = True

    # Compute average ROC
    base_fpr = np.linspace(0, 1, 1000)
    # base_fpr = np.logspace(-4, 0, 1000)
    all_tpr = np.zeros((len(metrics), len(base_fpr)), float)
    for i, metric_set in enumerate(metrics):
        all_tpr[i, :] = np.interp(base_fpr, metric_set["fpr"], metric_set["tpr"])

    if do_dummy:
        all_tpr_dummy = np.zeros((len(dummy_metrics), len(base_fpr)), float)
        for i, metric_set in enumerate(dummy_metrics):
            all_tpr_dummy[i, :] = np.interp(
                base_fpr, metric_set["fpr"], metric_set["tpr"]
            )

        for _, metric_set in enumerate(dummy_metrics):
            plt.plot(
                metric_set["fpr"],
                metric_set["tpr"],
                color="lightsteelblue",
                linewidth=0.5,
                alpha=0.5,
            )

    for _, metric_set in enumerate(metrics):
        plt.plot(
            metric_set["fpr"], metric_set["tpr"], color="lightsalmon", linewidth=0.5
        )

    tpr_mu = all_tpr.mean(axis=0)
    plt.plot(base_fpr, tpr_mu, "r")

    if do_dummy:
        dummy_mu = all_tpr_dummy.mean(axis=0)
        plt.plot(base_fpr, dummy_mu, "b")

    plt.xscale("log")
    plt.yscale("log")
    plt.xlabel("False Positive Rate")
    plt.ylabel("True Positive Rate")
    plt.tight_layout()
    plt.grid()
    plt.savefig(save_name)


def create_mia_report(attack_output: dict) -> FPDF:
    """make a worst case membership inference report

    Parameters
    ----------

    metadata: dict
        dictionary of metadata

    mia_metrics: list
        list of metrics dictionaries

    dummy_metrics: list
        list of metrics dictionaries

    dummy_metadata: dict
        metadata for dummy experiments

    Returns
    -------

    pdf: fpdf.FPDF
        fpdf document object

    """

    dummy_metrics = attack_output["dummy_attack_metrics"]
    mia_metrics = attack_output["attack_metrics"]
    metadata = attack_output["metadata"]
    if dummy_metrics is None or len(dummy_metrics) == 0:
        do_dummy = False
    else:
        do_dummy = True

    _roc_plot(mia_metrics, dummy_metrics, "log_roc.png")

    pdf = FPDF()
    pdf.add_page()
    pdf.set_xy(0, 0)
    title(pdf, "WorstCase MIA attack result report")
    subtitle(pdf, "Introduction")
    line(pdf, INTRODUCTION)
    subtitle(pdf, "Experiment summary")
    for key, value in metadata["experiment_details"].items():
        line(pdf, f"{key:>30s}: {str(value):30s}", font="courier")
    subtitle(pdf, "Global metrics")
    for key, value in metadata["global_metrics"].items():
        line(pdf, f"{key:>30s}: {str(value):30s}", font="courier")
    if do_dummy:
        subtitle(pdf, "Baseline global metrics")
        for key, value in metadata["baseline_global_metrics"].items():
            line(pdf, f"{key:>30s}: {str(value):30s}", font="courier")

    subtitle(pdf, "Metrics")
    line(
        pdf,
        "The following show summaries of the attack metrics over the repetitions",
        font="arial",
    )
    for metric in DISPLAY_METRICS:
        vals = np.array([m[metric] for m in mia_metrics])
        if metric in MAPPINGS:
            vals = np.array([MAPPINGS[metric](v) for v in vals])
        text = (
            f"{metric:>12} mean = {vals.mean():.2f}, var = {vals.var():.4f}, "
            f"min = {vals.min():.2f}, max = {vals.max():.2f}"
        )
        line(pdf, text, font="courier")

    if do_dummy:
        subtitle(pdf, "Baseline metrics")
        line(
            pdf,
            (
                "The following show summaries of the attack metrics over the "
                "repetitions where there is no statistical difference between "
                "predictions in the training and test sets. Simulation was done "
                "with training and test set sizes equal to the real ones"
            ),
            font="arial",
        )
        for metric in DISPLAY_METRICS:
            vals = np.array([m[metric] for m in dummy_metrics])
            if metric in MAPPINGS:
                vals = np.array([MAPPINGS[metric](v) for v in vals])
            text = (
                f"{metric:>12} mean = {vals.mean():.2f}, var = {vals.var():.4f}, "
                f"min = {vals.min():.2f}, max = {vals.max():.2f}"
            )
            line(pdf, text, font="courier")

    pdf.add_page()
    subtitle(pdf, "Log ROC")
    pdf.image("log_roc.png", x=None, y=None, w=0, h=140, type="", link="")
    pdf.set_font("arial", "", 12)
    line(pdf, LOGROC_CAPTION)

    pdf.add_page()
    title(pdf, "Glossary")
    _write_dict(pdf, GLOSSARY)

    return pdf


def create_json_report(output):
    """Create a report in json format for injestion by other tools"""
    # Initial work, just dump mia_metrics and dummy_metrics into a json structure
    return json.dumps(output, cls=NumpyArrayEncoder)


def create_lr_report(output):
    """TODO"""
    mia_metrics = output["attack_metrics"][0]
    metadata = output["metadata"]
    _roc_plot_single(mia_metrics, "log_roc.png")
    pdf = FPDF()
    pdf.add_page()
    pdf.set_xy(0, 0)
    title(pdf, "Likelihood Ratio Attack Report")
    subtitle(pdf, "Introduction")
    subtitle(pdf, "Metadata")
    for key, value in metadata["experiment_details"].items():
        line(pdf, f"{key:>30s}: {str(value):30s}", font="courier")
    for key, value in metadata["global_metrics"].items():
        line(pdf, f"{key:>30s}: {str(value):30s}", font="courier")
    subtitle(pdf, "Metrics")
    sub_metrics_dict = {
        key: val for key, val in mia_metrics.items() if isinstance(val, float)
    }
    for key, value in sub_metrics_dict.items():
        if key in MAPPINGS:
            value = MAPPINGS[key](value)
        line(pdf, f"{key:>30s}: {value:.4f}", font="courier")
    pdf.add_page()
    subtitle(pdf, "ROC Curve")
    pdf.image("log_roc.png", x=None, y=None, w=0, h=140, type="", link="")
    return pdf



"""test_loaders.py
Series of functions to use with pytest to check the loaders classes
Most use just the truncated files with first five examples of each class for brevity.
Please access the datasets from the sources listed in preprocessing/loaders.py
Please acknowledge those sources in any publications.
Jim Smith 2022
"""
import os
import shutil

import pytest

from aisdc.preprocessing import loaders
from aisdc.preprocessing.loaders import DataNotAvailable, UnknownDataset

PROJECT_ROOT_FOLDER = os.path.dirname(os.path.dirname(__file__))
DATA_FOLDER = os.path.join(PROJECT_ROOT_FOLDER, "tests", "datasets")

datasets = (
    "mimic2-iaccd",
    "in-hospital-mortality",
    "medical-mnist-ab-v-br-100",
    "medical-mnist-ab-v-br-500",
    "medical-mnist-all-100",
    "indian liver",
    "texas hospitals 10",
    "synth-ae",
    "synth-ae-small",
    "nursery",
    "iris",
)

preprocessing = ("standard", "minmax", "round")


def test_get_sklearn_dataset():
    """Test ability to load some standard datasets
    These loaders only return binary versions
    """
    # test preprocessing with iris for speed
    # just gets first two classes
    for routine in preprocessing:
        x_df, y_df = loaders.get_data_sklearn(routine + " iris")
        assert x_df.shape == (100, 4)
        assert y_df.shape == (100, 1)

    # iris without pre-processing
    x_df, y_df = loaders.get_data_sklearn("iris")
    assert x_df.shape == (100, 4)
    assert y_df.shape == (100, 1)

    # nursery
    x_df, y_df = loaders.get_data_sklearn("nursery")
    assert x_df.shape == (12960, 27)
    assert y_df.shape == (12960, 1)


def test_data_absent():
    """Tests exceptions raised when datasets have not been downloaded"""
    # mimic2
    with pytest.raises(DataNotAvailable):
        _, _ = loaders.get_data_sklearn("mimic2-iaccd")

    # in house mortality
    with pytest.raises(DataNotAvailable):
        _, _ = loaders.get_data_sklearn("in-hospital-mortality")

    # mnist 100
    with pytest.raises(DataNotAvailable):
        _, _ = loaders.get_data_sklearn("medical-mnist-ab-v-br-100")

    # mnist 500
    with pytest.raises(DataNotAvailable):
        _, _ = loaders.get_data_sklearn("medical-mnist-ab-v-br-500")

    # mnist all 100
    with pytest.raises(DataNotAvailable):
        _, _ = loaders.get_data_sklearn("medical-mnist-all-100")

    # indian liver
    with pytest.raises(DataNotAvailable):
        _, _ = loaders.get_data_sklearn("indian liver")

    # texas
    with pytest.raises(DataNotAvailable):
        _, _ = loaders.get_data_sklearn("texas hospitals 10")

    # synth-ae
    with pytest.raises(DataNotAvailable):
        _, _ = loaders.get_data_sklearn("synth-ae")

    # synthae-small
    with pytest.raises(DataNotAvailable):
        _, _ = loaders.get_data_sklearn("synth-ae-small")

    # unknown
    with pytest.raises(UnknownDataset):
        _, _ = loaders.get_data_sklearn("no such dataset")


def test_mimic():
    """load the mimic2 dataset"""
    x_df, y_df = loaders.get_data_sklearn("mimic2-iaccd", DATA_FOLDER)
    assert x_df.shape == (1064, 38), f"x_df shape is {x_df.shape}"
    assert y_df.shape == (1064, 1)


def test_in_hospital():
    """tests loading the in hospital mortality data
    in two different ways"""

    zip_file_name = os.path.join(DATA_FOLDER, "doi_10.5061_dryad.0p2ngf1zd__v5.zip")
    new_file_name = os.path.join(DATA_FOLDER, "doi_10.5061_dryad.0p2ngf1zd__v5.renamed")
    # first attempt reads from zip file

    x_df, y_df = loaders.get_data_sklearn("in-hospital-mortality", DATA_FOLDER)
    assert x_df.shape == (428, 48), f"x_df shape is {x_df.shape}"
    assert y_df.shape == (428, 1)

    # now move zip file and make sure the loader reads properly from "data01.csv"
    os.rename(zip_file_name, new_file_name)
    x_df, y_df = loaders.get_data_sklearn("in-hospital-mortality", DATA_FOLDER)
    assert x_df.shape == (428, 48), f"x_df shape is {x_df.shape}"
    assert y_df.shape == (428, 1)

    # then put the zip file back for next time we run this test
    os.rename(new_file_name, zip_file_name)


def test_mnist():
    """tests loading medical mnist data from different sources"""
    # mnist 100
    # this ne assumes the zip file is present
    x_df, y_df = loaders.get_data_sklearn("medical-mnist-ab-v-br-100", DATA_FOLDER)
    assert x_df.shape == (10, 4096), f"x_df shape is {x_df.shape}"
    assert y_df.shape == (10, 1)

    # mnist 500
    # now the archive file is present
    x_df, y_df = loaders.get_data_sklearn("medical-mnist-ab-v-br-500", DATA_FOLDER)
    assert x_df.shape == (10, 4096), f"x_df shape is {x_df.shape}"
    assert y_df.shape == (10, 1)

    # mnist all 100
    x_df, y_df = loaders.get_data_sklearn("medical-mnist-all-100", DATA_FOLDER)
    assert x_df.shape == (30, 4096), f"x_df shape is {x_df.shape}"
    assert y_df.shape == (30, 1)

    # test unflattend images
    imgdir = os.path.join(DATA_FOLDER, "kaggle-medical-mnist", "archive", "CXR")
    x_df, y_df = loaders._images_to_ndarray(  # pylint:disable=protected-access
        imgdir, 5, 0, flatten=False
    )
    assert x_df.shape == (5, 64, 64), f"x_df shape is {x_df.shape}"
    assert y_df.shape == (5, 1)

    # remove archive folder
    name = os.path.join(DATA_FOLDER, "kaggle-medical-mnist", "archive")
    if os.path.exists(name) and os.path.isdir(name):
        shutil.rmtree(name)


def test_indian_liver():
    """the indian liver dataloader"""
    x_df, y_df = loaders.get_data_sklearn("indian liver", DATA_FOLDER)
    assert x_df.shape == (11, 10), f"x_df shape is {x_df.shape}"
    assert y_df.shape == (11, 1)


def test_texas():
    """the texas dataset has quite restrictive licenses arounbd copying.
    Therefore it is omitted from the CI testing.
    This code was fully tested during the GRAIMatter project
    """


def test_synth_ae():
    """tests different versions of the  synthetic A&E dataset"""
    x_df, y_df = loaders.get_data_sklearn("synth-ae", DATA_FOLDER)
    assert x_df.shape == (8, 16), f"x_df shape is {x_df.shape}"
    assert y_df.shape == (8, 1)

    # synthae-small
    x_df, y_df = loaders.get_data_sklearn("synth-ae-small", DATA_FOLDER)
    assert x_df.shape == (8, 16), f"x_df shape is {x_df.shape}"
    assert y_df.shape == (8, 1)



"""This module contains unit tests for the SafeDecisionTreeClassifier."""

import os
import pickle

import joblib
import numpy as np
from sklearn import datasets

from aisdc.safemodel import reporting
from aisdc.safemodel.classifiers import SafeDecisionTreeClassifier
from aisdc.safemodel.classifiers.safedecisiontreeclassifier import (
    decision_trees_are_equal,
    get_tree_k_anonymity,
)


def get_data():
    """Returns data for testing."""
    iris = datasets.load_iris()
    x = np.asarray(iris["data"], dtype=np.float64)
    y = np.asarray(iris["target"], dtype=np.float64)
    x = np.vstack([x, (7, 2.0, 4.5, 1)])
    y = np.append(y, 4)
    return x, y


def test_reporting():
    """Check that getting report name requires name"""
    correct_msg = "Error - get_reporting_string: No 'name' given"
    msg = reporting.get_reporting_string()
    assert msg == correct_msg


def test_decision_trees_are_equal():
    """test the code that compares two decision trees"""
    x1, y = get_data()
    model1 = SafeDecisionTreeClassifier(random_state=1)
    model1.fit(x1, y)

    # same
    model2 = SafeDecisionTreeClassifier(random_state=1)
    model2.fit(x1, y)
    same, _ = decision_trees_are_equal(model1, model2)

    # one or both untrained
    model3 = SafeDecisionTreeClassifier(random_state=1, max_depth=7)
    same, msg = decision_trees_are_equal(model1, model3)
    assert same is False
    assert len(msg) > 0
    same, msg = decision_trees_are_equal(model3, model1)
    assert same is False
    assert len(msg) > 0
    same, msg = decision_trees_are_equal(model3, model3)
    assert same is True
    # assert len(msg)>0

    # different
    # x2=x1+1
    model3 = SafeDecisionTreeClassifier(random_state=1, max_depth=7)
    model3.criterion = "entropy"
    model3.fit(x1, y)
    same, msg = decision_trees_are_equal(model1, model3)
    print(f"diff msg = {msg}")
    assert same is False

    # wrong type
    same, _ = decision_trees_are_equal(model1, "aString")
    assert same is False


def test_get_tree_k_anonymity():
    """getting k_anonymity
    50 data points randomly split in 2, single layer
    so k should be ~25
    """
    x = np.random.rand(50, 2)
    y = np.ones(50)
    for i in range(x.shape[0]):
        if x[i][0] < 0.5:
            y[i] = 0
    model = SafeDecisionTreeClassifier(random_state=1, max_depth=1)
    model.fit(x, y)
    k = get_tree_k_anonymity(model, x)
    # print(f'k={k}')
    assert k > 10


def test_decisiontree_unchanged():
    """SafeDecisionTreeClassifier using unchanged values."""
    x, y = get_data()
    model = SafeDecisionTreeClassifier(random_state=1)
    model.fit(x, y)
    assert model.score(x, y) == 0.9668874172185431
    msg, disclosive = model.preliminary_check()
    correct_msg = "Model parameters are within recommended ranges.\n"
    assert msg == correct_msg
    assert disclosive is False


def test_decisiontree_safe_recommended():
    """SafeDecisionTreeClassifier using recommended values."""
    x, y = get_data()
    model = SafeDecisionTreeClassifier(random_state=1)
    model.min_samples_leaf = 5
    model.fit(x, y)
    assert model.score(x, y) == 0.9668874172185431
    msg, disclosive = model.preliminary_check()
    correct_msg = "Model parameters are within recommended ranges.\n"
    assert msg == correct_msg
    assert disclosive is False


def test_decisiontree_safe_1():
    """SafeDecisionTreeClassifier with safe changes."""
    x, y = get_data()
    model = SafeDecisionTreeClassifier(random_state=1)
    model.min_samples_leaf = 10
    model.fit(x, y)
    assert model.score(x, y) == 0.9536423841059603
    msg, disclosive = model.preliminary_check()
    correct_msg = "Model parameters are within recommended ranges.\n"
    assert msg == correct_msg
    assert disclosive is False


def test_decisiontree_safe_2():
    """SafeDecisionTreeClassifier with safe changes."""
    x, y = get_data()
    model = SafeDecisionTreeClassifier(random_state=1, min_samples_leaf=10)
    model.fit(x, y)
    assert model.score(x, y) == 0.9536423841059603
    msg, disclosive = model.preliminary_check()
    correct_msg = "Model parameters are within recommended ranges.\n"
    assert msg == correct_msg
    assert disclosive is False


def test_decisiontree_unsafe_1():
    """SafeDecisionTreeClassifier with unsafe changes."""
    x, y = get_data()
    model = SafeDecisionTreeClassifier(random_state=1)
    model.min_samples_leaf = 1
    model.fit(x, y)
    assert model.score(x, y) == 1
    msg, disclosive = model.preliminary_check()
    correct_msg = (
        "WARNING: model parameters may present a disclosure risk:\n"
        "- parameter min_samples_leaf = 1 identified as less than the recommended "
        "min value of 5."
    )
    assert msg == correct_msg
    assert disclosive is True


def test_decisiontree_unsafe_2():
    """SafeDecisionTreeClassifier with unsafe changes - automatically fixed."""
    x, y = get_data()
    model = SafeDecisionTreeClassifier(random_state=1, min_samples_leaf=1)
    model.fit(x, y)
    assert model.score(x, y) == 0.9668874172185431
    msg, disclosive = model.preliminary_check()
    correct_msg = "Model parameters are within recommended ranges.\n"
    assert msg == correct_msg
    assert disclosive is False


def test_decisiontree_save():
    """SafeDecisionTreeClassifier model saving."""
    x, y = get_data()
    model = SafeDecisionTreeClassifier(random_state=1, min_samples_leaf=50)
    model.fit(x, y)
    assert model.score(x, y) == 0.9470198675496688
    # test pickle
    model.save("dt_test.pkl")
    with open("dt_test.pkl", "rb") as file:
        pkl_model = pickle.load(file)
    assert pkl_model.score(x, y) == 0.9470198675496688
    # test joblib
    model.save("dt_test.sav")
    with open("dt_test.sav", "rb") as file:
        sav_model = joblib.load(file)
    assert sav_model.score(x, y) == 0.9470198675496688

    # cleanup
    for name in ("dt_test.pkl", "dt_test.sav"):
        if os.path.exists(name) and os.path.isfile(name):
            os.remove(name)


def test_decisiontree_hacked_postfit():
    """SafeDecisionTreeClassifier changes made to parameters after fit() called."""
    x, y = get_data()
    model = SafeDecisionTreeClassifier(random_state=1, min_samples_leaf=1)
    model.min_samples_leaf = 1
    model.fit(x, y)
    assert model.score(x, y) == 1.0
    model.min_samples_leaf = 5

    # forcing errors in internal tree checking by hacking tree
    model.tree_.max_depth = 5
    model.tree_.value[0][-1] = 1

    # At first glance the model now looks ok
    msg1, disclosive = model.preliminary_check()
    correct_msg1 = reporting.get_reporting_string(name="within_recommended_ranges")
    print(f"correct msg1: {correct_msg1}\n", f"actual  msg1: {msg1}")
    assert msg1 == correct_msg1
    assert disclosive is False

    # but on closer inspection not
    msg2, disclosive2 = model.posthoc_check()
    part1 = reporting.get_reporting_string(name="basic_params_differ", length=1)
    part2 = "parameter min_samples_leaf changed from 1 to 5 after model was fitted.\n"
    part3 = reporting.get_reporting_string(
        name="internal_attribute_differs", attr="max_depth"
    )
    part4 = reporting.get_reporting_string(
        name="internal_attribute_differs", attr="value"
    )

    correct_msg2 = part1 + part2 + part3 + part4
    print(f"Correct msg2 : {correct_msg2}\n" f"actual mesg2 : {msg2}")
    assert msg2 == correct_msg2
    assert disclosive2 is True


def test_data_hiding():
    """what if the hacking was really obscure
    like putting something in the exceptions list
     then adding data to current and saved copies"""
    x, y = get_data()
    model = SafeDecisionTreeClassifier(random_state=1, min_samples_leaf=5)
    model.fit(x, y)
    # now the hack
    model.examine_seperately_items = ["tree_", "bad_string"]
    model.bad_string = "something disclosive"
    model.saved_model["bad_string"] = "something disclosive"
    part1 = reporting.get_reporting_string(name="basic_params_differ", length=1)
    part2 = "('remove', 'examine_seperately_items', [(1, 'bad_string')])"
    part3 = reporting.get_reporting_string(name="unexpected_item")
    correct_msg = part1 + part2 + part3
    msg, disclosive = model.posthoc_check()
    print(f"Correct msg : {correct_msg}\n" f"actual  msg : {msg}")
    assert disclosive
    assert msg == correct_msg



"""dataset.py - class to represent a dataset"""

from __future__ import annotations

import numpy as np


class Data:  # pylint: disable=too-many-instance-attributes
    """Stripped down Data class"""

    def __init__(self) -> None:
        self.name: str = ""
        self.n_samples: int = 0
        self.x_train: np.ndarray
        self.y_train: np.ndarray
        self.x_test: np.ndarray
        self.y_test: np.ndarray

        self.features: dict = {}
        self.n_features: int = 0
        self.x_orig: np.ndarray
        self.y_orig: np.ndarray
        self.x_train_orig: np.ndarray
        self.y_train_orig: np.ndarray
        self.x_test_orig: np.ndarray
        self.y_test_orig: np.ndarray
        self.n_samples_orig: int = 0

    def add_processed_data(
        self,
        x_train: np.ndarray,
        y_train: np.ndarray,
        x_test: np.ndarray,
        y_test: np.ndarray,
    ) -> None:
        """Add a processed and split dataset"""
        self.x_train = x_train
        self.y_train = np.array(y_train, int)
        self.x_test = x_test
        self.y_test = np.array(y_test, int)
        self.n_samples = len(x_train) + len(y_train)

    def add_feature(self, name: str, indices: list[int], encoding: str) -> None:
        """Add a feature description to the data dictionary."""
        index: int = len(self.features)
        self.features[index] = {
            "name": name,
            "indices": indices,
            "encoding": encoding,
        }
        self.n_features = len(self.features)

    def add_raw_data(  # pylint: disable=too-many-arguments
        self,
        x_orig: np.ndarray,
        y_orig: np.ndarray,
        x_train_orig: np.ndarray,
        y_train_orig: np.ndarray,
        x_test_orig: np.ndarray,
        y_test_orig: np.ndarray,
    ) -> None:
        """Add original unprocessed dataset"""
        self.x_orig = x_orig
        self.y_orig = y_orig
        self.x_train_orig = x_train_orig
        self.y_train_orig = y_train_orig
        self.x_test_orig = x_test_orig
        self.y_test_orig = y_test_orig
        self.n_samples_orig = len(x_orig)

    def __str__(self):
        return self.name



"""
Test the metrics
"""

import unittest

import numpy as np

from aisdc.attacks.metrics import _div, _tpr_at_fpr, get_metrics
from aisdc.attacks.mia_extremecase import min_max_disc

# pylint: disable = invalid-name

PREDICTED_CLASS = np.array([0, 1, 0, 0, 1, 1])
TRUE_CLASS = np.array([0, 0, 0, 1, 1, 1])
PREDICTED_PROBS = np.array(
    [[0.9, 0.1], [0.4, 0.6], [0.8, 0.2], [0.55, 0.45], [0.1, 0.9], [0.01, 0.99]]
)


class DummyClassifier:
    """
    Mocks the predict and predict_proba methods
    """

    def predict(self, _):
        """
        return dummy predictions
        """
        return PREDICTED_CLASS

    def predict_proba(self, _):
        """
        return dummy predicted probabilities
        """
        return PREDICTED_PROBS


class TestMetrics(unittest.TestCase):
    """
    Test the metrics with some dummy predictions
    """

    def test_metrics(self):
        """
        Test each individual metric with dummy data
        """
        clf = DummyClassifier()
        testX = []
        testy = TRUE_CLASS
        metrics = get_metrics(clf, testX, testy, permute_rows=False)
        self.assertAlmostEqual(metrics["TPR"], 2 / 3)
        self.assertAlmostEqual(metrics["FPR"], 1 / 3)
        self.assertAlmostEqual(metrics["FAR"], 1 / 3)
        self.assertAlmostEqual(metrics["TNR"], 2 / 3)
        self.assertAlmostEqual(metrics["PPV"], 2 / 3)
        self.assertAlmostEqual(metrics["NPV"], 2 / 3)
        self.assertAlmostEqual(metrics["FNR"], 1 / 3)
        self.assertAlmostEqual(metrics["ACC"], 4 / 6)
        self.assertAlmostEqual(metrics["F1score"], (8 / 9) / (2 / 3 + 2 / 3))
        self.assertAlmostEqual(metrics["Advantage"], 1 / 3)
        self.assertAlmostEqual(metrics["AUC"], 8 / 9)


class TestFPRatTPR(unittest.TestCase):
    """
    Test code that computes TPR at fixed FPR
    """

    def test_tpr(self):
        """Test tpr at fpr"""
        y_true = TRUE_CLASS
        y_score = PREDICTED_PROBS[:, 1]

        tpr = _tpr_at_fpr(y_true, y_score, fpr=0)
        self.assertAlmostEqual(tpr, 2 / 3)
        tpr = _tpr_at_fpr(y_true, y_score, fpr=0.001)
        self.assertAlmostEqual(tpr, 2 / 3)
        tpr = _tpr_at_fpr(y_true, y_score, fpr=0.1)
        self.assertAlmostEqual(tpr, 2 / 3)
        tpr = _tpr_at_fpr(y_true, y_score, fpr=0.4)
        self.assertAlmostEqual(tpr, 1)
        tpr = _tpr_at_fpr(y_true, y_score, fpr=1.0)
        self.assertAlmostEqual(tpr, 1)
        tpr = _tpr_at_fpr(y_true, y_score, fpr_perc=True, fpr=100.0)
        self.assertAlmostEqual(tpr, 1)


class Test_Div(unittest.TestCase):
    """tests the _div functionality"""

    def test_div(self):
        """test div for y=1 and 0"""
        result = _div(8.0, 1.0, 99.0)
        self.assertAlmostEqual(result, 8.0)
        result2 = _div(8.0, 0.0, 99.0)
        self.assertAlmostEqual(result2, 99.0)


class TestExtreme(unittest.TestCase):
    """
    Test the extreme metrics
    """

    def test_extreme_default(self):
        """
        Tets with the dummy data
        """
        pred_probs = DummyClassifier().predict_proba(None)[:, 1]
        maxd, mind, mmd, _ = min_max_disc(TRUE_CLASS, pred_probs)

        # 10% of 6 is 1 so:
        # maxd should be 1 (the highest one is predicted as1)
        # mind should be 0 (the lowest one is not predicted as1)
        self.assertAlmostEqual(maxd, 1.0)
        self.assertAlmostEqual(mind, 0.0)
        self.assertAlmostEqual(mmd, 1.0)

    def test_extreme_higer_prop(self):
        """
        Tets with the dummy data but increase proportion to 0.5
        """
        pred_probs = DummyClassifier().predict_proba(None)[:, 1]
        maxd, mind, mmd, _ = min_max_disc(TRUE_CLASS, pred_probs, x_prop=0.5)

        # 10% of 6 is 1 so:
        # maxd should be 1 (the highest one is predicted as1)
        # mind should be 0 (the lowest one is not predicted as1)
        self.assertAlmostEqual(maxd, 2 / 3)
        self.assertAlmostEqual(mind, 1 / 3)
        self.assertAlmostEqual(mmd, 1 / 3)



"""Python setup script for installing AI-SDC."""

from pathlib import Path

from setuptools import find_packages, setup

this_directory = Path(__file__).parent
long_description = (this_directory / "README.md").read_text()

setup(
    name="aisdc",
    version="1.0.2",
    license="MIT",
    maintainer="Jim Smith",
    maintainer_email="james.smith@uwe.ac.uk",
    description="Tools for the statistical disclosure control of machine learning models",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/AI-SDC/AI-SDC",
    packages=find_packages(exclude=["tests"]),
    package_data={"aisdc.safemodel": ["rules.json"]},
    python_requires=">=3.8,<3.11",
    install_requires=[
        "dictdiffer~=0.9.0",
        "fpdf~=1.7.2",
        "joblib~=1.1.0",
        "multiprocess~=0.70.12.2",
        "numpy~=1.21",
        "pandas~=1.5.0",
        "scikit_learn~=1.1.2",
        "scipy~=1.7",
        "tensorflow~=2.10.0",
        "tensorflow_privacy~=0.8.7",
    ],
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Intended Audience :: Developers",
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: MIT License",
        "Natural Language :: English",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Topic :: Scientific/Engineering",
        "Topic :: Scientific/Engineering :: Artificial Intelligence",
        "Topic :: Scientific/Engineering :: Information Analysis",
        "Operating System :: OS Independent",
    ],
    keywords=[
        "data-privacy",
        "data-protection",
        "machine-learning",
        "privacy",
        "privacy-tools",
        "statistical-disclosure-control",
    ],
)



"""Method for computing extreme case metrics"""

from __future__ import annotations

import numpy as np
from scipy.stats import norm


def min_max_disc(
    y_true: np.ndarray, pred_probs: np.ndarray, x_prop: float = 0.1, log_p: bool = True
) -> tuple[float, float, float, float]:  # pylint: disable = line-too-long
    """
    Non-average-case methods for MIA attacks. Considers actual frequency of membership
    amongst samples with highest- and lowest- assessed probability of membership. If an
    MIA method confidently asserts that 5% of samples are members and 5% of samples are
    not, but cannot tell for the remaining 90% of samples, then these metrics will flag
    this behaviour, but AUC/advantage may not. Since the difference may be noisy, a
    p-value against a null of independence of true membership and assessed membership
    probability (that is, membership probabilities are essentially random) is also used
    as a metric (using a usual Gaussian approximation to binomial). If the p-value is
    low and the frequency difference is high (>0.5) then the MIA attack is successful
    for some samples.

    Parameters
    ----------
        y: np.ndarray
            true labels
        yp: np.ndarray
            probabilities of labels, or monotonic transformation of probabilties
        xprop: float
            proportion of samples with highest- and lowest- probability of membership to be
            considered
        logp: bool
            convert p-values to log(p).

    Returns
    -------
        maxd: float
            frequency of y=1 amongst proportion xprop of individuals with highest assessed
            membership probability
        mind: float
            frequency of y=1 amongst proportion xprop of individuals with lowest assessed
            membership probability
        mmd: float
            difference between maxd and mind
        pval: float
            p-value or log-p value corresponding to mmd against null hypothesis that random
            variables corresponding to y and yp are independent.

    Notes
    -----

    Examples
    --------
    >>> y = np.random.choice(2, 100)
    >>> yp = np.random.rand(100)
    >>> maxd, mind, mmd, pval = min_max_desc(y, yp, xprop=0.2, logp=True)

    """

    n_examples = int(np.ceil(len(y_true) * x_prop))
    pos_frequency = np.mean(y_true)  # average frequency
    y_order = np.argsort(pred_probs)  # ordering permutation

    # Frequencies
    # y values corresponding to lowest k values of yp
    y_lowest_n = y_true[y_order[:n_examples]]
    # y values corresponding to highest k values of yp
    y_highest_n = y_true[y_order[-(n_examples):]]
    maxd = np.mean(y_highest_n)
    mind = np.mean(y_lowest_n)
    mmd = maxd - mind

    # P-value
    # mmd is asymptotically distributed as N(0,sdm^2) under null.
    sdm = np.sqrt(2 * pos_frequency * (1 - pos_frequency) / n_examples)
    pval = 1 - norm.cdf(mmd, loc=0, scale=sdm)  # normal CDF
    if log_p:
        if pval < 1e-50:
            pval = -115.13
        else:
            pval = np.log(pval)

    # Return
    return maxd, mind, mmd, pval



""" code to test the file attacks/dataset
Jim Smith james.smith@uwe.ac.uk 2022
"""
import numpy as np
from sklearn.datasets import fetch_openml
from sklearn.model_selection import train_test_split
from sklearn.preprocessing import LabelEncoder, OneHotEncoder

from aisdc.attacks.dataset import Data


def test_dataset():  # pylint:disable=too-many-locals
    """returns a randomly sampled 10+10% of
    the nursery data set as a Data object
    if needed fetches it from openml and saves. it

    """

    nursery_data = fetch_openml(data_id=26, as_frame=True)
    x = np.asarray(nursery_data.data, dtype=str)
    y = np.asarray(nursery_data.target, dtype=str)
    # change labels from recommend to priority for the two odd cases
    num = len(y)
    for i in range(num):
        if y[i] == "recommend":
            y[i] = "priority"

    indices: list[list[int]] = [
        [0, 1, 2],  # parents
        [3, 4, 5, 6, 7],  # has_nurs
        [8, 9, 10, 11],  # form
        [12, 13, 14, 15],  # children
        [16, 17, 18],  # housing
        [19, 20],  # finance
        [21, 22, 23],  # social
        [24, 25, 26],  # health
        [27],  # dummy
    ]

    # [Researcher] Split into training and test sets
    # target model train / test split - these are strings
    (
        x_train_orig,
        x_test_orig,
        y_train_orig,
        y_test_orig,
    ) = train_test_split(
        x,
        y,
        test_size=0.05,
        stratify=y,
        shuffle=True,
    )

    # now resample the training data reduce number of examples
    _, x_train_orig, _, y_train_orig = train_test_split(
        x_train_orig,
        y_train_orig,
        test_size=0.05,
        stratify=y_train_orig,
        shuffle=True,
    )

    # [Researcher] Preprocess dataset
    # one-hot encoding of features and integer encoding of labels
    label_enc = LabelEncoder()
    feature_enc = OneHotEncoder()
    x_train = feature_enc.fit_transform(x_train_orig).toarray()
    y_train = label_enc.fit_transform(y_train_orig)
    x_test = feature_enc.transform(x_test_orig).toarray()
    y_test = label_enc.transform(y_test_orig)

    # add dummy continuous valued attribute
    dummy_tr = np.random.rand(x_train.shape[0], 1)
    dummy_te = np.random.rand(x_test.shape[0], 1)
    x_train = np.hstack((x_train, dummy_tr))
    x_train_orig = np.hstack((x_train_orig, dummy_tr))
    x_test = np.hstack((x_test, dummy_te))
    x_test_orig = np.hstack((x_test_orig, dummy_te))

    n_features = np.shape(x_train_orig)[1]

    # [TRE / Researcher] Wrap the data in a dataset object
    the_data = Data()
    the_data.name = "nursery"
    the_data.add_processed_data(x_train, y_train, x_test, y_test)
    the_data.add_raw_data(x, y, x_train_orig, y_train_orig, x_test_orig, y_test_orig)
    for i in range(n_features - 1):
        the_data.add_feature(nursery_data.feature_names[i], indices[i], "onehot")
    the_data.add_feature("dummy", indices[n_features - 1], "float")



"""This module contains unit tests for SafeKerasModel."""

import getpass
import os
import platform
import shutil

# shut tensorflow up
import warnings

import numpy as np
import tensorflow as tf
from sklearn import datasets
from sklearn.model_selection import train_test_split
from tensorflow.keras.layers import (  # pylint: disable = import-error, no-name-in-module
    Dense,
    Input,
)

from aisdc.safemodel.classifiers import SafeKerasModel, safekeras
from aisdc.safemodel.reporting import get_reporting_string

warnings.filterwarnings("ignore", category=DeprecationWarning)
warnings.filterwarnings("ignore", category=FutureWarning)


EPOCHS = 1
n_classes = 4
# expected accuracy
ACC = 0.325 if platform.system() == "Darwin" else 0.3583333492279053

UNSAFE_ACC = 0.325 if platform.system() == "Darwin" else 0.3583333492279053


def cleanup_file(name: str):
    """removes unwanted files or directory"""
    if os.path.exists(name) and os.path.isfile(name):  # h5
        os.remove(name)
    elif os.path.exists(name) and os.path.isdir(name):  # tf
        shutil.rmtree(name)


def get_data():
    """Returns data for testing."""
    iris = datasets.load_iris()
    xall = np.asarray(iris["data"], dtype=np.float64)
    yall = np.asarray(iris["target"], dtype=np.float64)
    xall = np.vstack([xall, (7, 2.0, 4.5, 1)])
    yall = np.append(yall, n_classes)
    X, Xval, y, yval = train_test_split(
        xall, yall, test_size=0.2, shuffle=True, random_state=12345
    )
    y = tf.one_hot(y, n_classes)
    yval = tf.one_hot(yval, n_classes)
    return X, y, Xval, yval


def make_small_model(num_hidden_layers=2):
    "Make the keras model"
    # get data
    X, y, Xval, yval = get_data()
    # set seed and kernel initialisers for repeatability
    tf.random.set_seed(12345)
    initializer = tf.keras.initializers.Zeros()
    input_data = Input(shape=X[0].shape)
    xx = Dense(32, activation="relu", kernel_initializer=initializer)(input_data)
    layers_added = 1
    while layers_added < num_hidden_layers:
        xx = Dense(32, activation="relu", kernel_initializer=initializer)(xx)
        layers_added += 1
    output = Dense(n_classes, activation="softmax", kernel_initializer=initializer)(xx)
    model = SafeKerasModel(
        inputs=input_data,
        outputs=output,
        name="test",
        num_samples=X.shape[0],
        epochs=EPOCHS,
    )

    return model, X, y, Xval, yval


def check_init_completed(model: SafeKerasModel):
    """basic checks for things that happen
    at end of correct init
    """
    rightname = "KerasModel"
    assert (
        model.model_type == rightname
    ), "failed check for model type being set in init()"
    errstr = (
        "default value of noise_multiplier=0.5"
        "Should have been reset to value 0.7 in rules.json"
    )
    assert model.noise_multiplier == 0.7, errstr


def test_init_variants():
    """Test alternative ways of calling init
    do just with single layer for speed of testing
    """
    # get data
    X, _, _, _ = get_data()
    # set seed and kernel initialisers for repeatability
    tf.random.set_seed(12345)
    initializer = tf.keras.initializers.Zeros()

    # define model architecture
    input_data = Input(shape=X[0].shape)
    xx = Dense(32, activation="relu", kernel_initializer=initializer)(input_data)
    output = Dense(n_classes, activation="softmax", kernel_initializer=initializer)(xx)

    # standard way
    model = SafeKerasModel(
        inputs=input_data,
        outputs=output,
        name="test",
        num_samples=X.shape[0],
        epochs=EPOCHS,
    )
    check_init_completed(model)

    # inputs and outputs not in kwargs
    model2 = SafeKerasModel(
        input_data,
        output,
        "test",
        num_samples=X.shape[0],
        epochs=EPOCHS,
    )
    check_init_completed(model2)

    # batch size zero
    model3 = SafeKerasModel(
        inputs=input_data,
        outputs=output,
        name="test",
        num_samples=X.shape[0],
        epochs=EPOCHS,
        batch_size=0,
    )
    errstr = "failed to correct batch_size=0 in init"
    assert model3.batch_size == 32, errstr


def test_same_configs():  # pylint: disable=too-many-locals
    """check whether tests for equal configuration work"""

    model1, X, _, _, _ = make_small_model(num_hidden_layers=1)
    model2, _, _, _, _ = make_small_model(num_hidden_layers=2)
    model2a, _, _, _, _ = make_small_model(num_hidden_layers=2)

    # different numbers of layers
    same1, msg1 = safekeras.same_configs(model1, model2)
    errstr = (
        f"model1 has {len(model1.layers)} layers, but\n"
        f"model2 has {len(model2.layers)} layers\n"
    )
    assert same1 is False, errstr
    correct_msg1 = get_reporting_string(name="different_layer_count")
    errstr = f"msg    was: {msg1}\n" f"should be : {correct_msg1}"
    assert msg1 == correct_msg1, errstr

    # same architecture
    same2, msg2 = safekeras.same_configs(model2, model2a)
    correct_msg2 = get_reporting_string(name="same_ann_config")
    assert msg2 == correct_msg2, (
        rf"should report {correct_msg2}\," f" but   got    {msg2}.\n"
    )
    assert same2 is True, "models are same!"

    # same layers, different widths
    input_data = Input(shape=X[0].shape)
    initializer = tf.keras.initializers.Zeros()
    xx = Dense(64, activation="relu", kernel_initializer=initializer)(input_data)
    output = Dense(n_classes, activation="softmax", kernel_initializer=initializer)(xx)
    model1a = SafeKerasModel(
        inputs=input_data,
        outputs=output,
        name="test",
        num_samples=X.shape[0],
        epochs=EPOCHS,
    )
    check_init_completed(model1a)
    same3, msg3 = safekeras.same_configs(model1, model1a)
    errmsg = "Should report layers have different num nodes"
    assert same3 is False, errmsg
    correct_msg3 = get_reporting_string(name="layer_configs_differ", layer=1, length=1)
    correct_msg3 += get_reporting_string(
        name="param_changed_from_to", key="units", val="32", cur_val=64
    )
    errstr = f"got message: {msg3}\n" rf"expected.  : {correct_msg3}\."
    assert msg3 == correct_msg3, errstr


def test_same_weights():  # pylint : disable=too-many-locals
    """check the same weights method catches differences"""
    # make models to test
    model1, X, _, _, _ = make_small_model(num_hidden_layers=1)
    model2, _, _, _, _ = make_small_model(num_hidden_layers=2)
    input_data = Input(shape=X[0].shape)
    initializer = tf.keras.initializers.Zeros()
    xx = Dense(64, activation="relu", kernel_initializer=initializer)(input_data)
    output = Dense(n_classes, activation="softmax", kernel_initializer=initializer)(xx)
    model1a = SafeKerasModel(
        inputs=input_data,
        outputs=output,
        name="test",
        num_samples=X.shape[0],
        epochs=EPOCHS,
    )

    # same
    same1, _ = safekeras.same_weights(model1, model1)
    assert same1 is True

    # different num layers
    same2, _ = safekeras.same_weights(model1, model2)
    assert same2 is False

    # different sized layers
    same3, _ = safekeras.same_weights(model1, model1a)
    errstr = (
        "model1 hidden layer has "
        f"  {len(model1.layers[1].get_weights()[0][0])} units"
        f" but model2 has {len(model1a.layers[1].get_weights()[0][0])}.\n"
    )
    assert same3 is False, errstr


def test_DP_optimizer_checks():
    """tests the various checks that DP optimiser was used"""
    # make model
    model1, _, _, _, _ = make_small_model(num_hidden_layers=1)
    loss = tf.keras.losses.CategoricalCrossentropy(
        from_logits=False, reduction=tf.losses.Reduction.NONE
    )

    # compile with allowed optimizer and test check works for positive case
    allowed_optimizers = [
        "tensorflow_privacy.DPKerasAdagradOptimizer",
        "tensorflow_privacy.DPKerasAdamOptimizer",
        "tensorflow_privacy.DPKerasSGDOptimizer",
    ]
    for oktype in allowed_optimizers:
        model, _, _, _, _ = make_small_model(num_hidden_layers=1)
        model.compile(loss=loss, optimizer=oktype)
        opt_ok, msg = safekeras.check_optimizer_allowed(model.optimizer)
        assert opt_ok is True, msg
        opt_is_dp, _ = safekeras.check_optimizer_is_DP(model.optimizer)
        assert opt_is_dp is True

    # not ok optimizer
    model, _, _, _, _ = make_small_model(num_hidden_layers=1)
    model.compile(loss=loss)
    # reset to not DP optimizer
    model.optimizer = tf.keras.optimizers.get("SGD")
    opt_ok, msg = safekeras.check_optimizer_allowed(model1.optimizer)
    assert opt_ok is False, msg
    opt_is_dp, msg = safekeras.check_optimizer_is_DP(model1.optimizer)
    assert opt_is_dp is False, msg


def test_DP_used():
    """tests the various checks that DP optimiser was used"""
    # should pass aftyer model compiled **and** fitted with DP optimizer
    model1, X, y, Xval, yval = make_small_model(num_hidden_layers=1)
    loss = tf.keras.losses.CategoricalCrossentropy(
        from_logits=False, reduction=tf.losses.Reduction.NONE
    )
    model1.compile(loss=loss)
    dp_used, msg = safekeras.check_DP_used(model1.optimizer)
    assert dp_used is False
    model1.fit(X, y, validation_data=(Xval, yval), epochs=EPOCHS, batch_size=20)
    dp_used, msg = safekeras.check_DP_used(model1.optimizer)
    assert dp_used is True

    # this model gets changed to non-DP by calling the superclass compile()
    # so should fail all the checks
    model2, _, _, _, _ = make_small_model(num_hidden_layers=1)
    super(SafeKerasModel, model2).compile(loss=loss, optimizer="SGD")
    model2.fit(X, y, validation_data=(Xval, yval), epochs=EPOCHS, batch_size=20)
    dp_used, msg = safekeras.check_DP_used(model2.optimizer)
    assert dp_used is False, msg


def test_checkpoints_are_equal():
    """test the check for checkpoint equality"""
    model1, X, y, Xval, yval = make_small_model(num_hidden_layers=1)
    loss = tf.keras.losses.CategoricalCrossentropy(
        from_logits=False, reduction=tf.losses.Reduction.NONE
    )
    model1.compile(loss=loss)
    model1.fit(X, y, validation_data=(Xval, yval), epochs=EPOCHS, batch_size=20)
    model1.save("fit.tf")
    model1.fit(X, y, validation_data=(Xval, yval), epochs=EPOCHS * 2, batch_size=20)
    model1.save("refit.tf")

    # same arch, different weights
    same, msg = safekeras.check_checkpoint_equality("fit.tf", "refit.tf")
    assert same is False, msg

    # should be same
    same, msg = safekeras.check_checkpoint_equality("fit.tf", "fit.tf")
    print(msg)
    assert same is True, msg

    # different architecture
    model2, X, y, Xval, yval = make_small_model(num_hidden_layers=3)
    model2.compile(loss=loss)
    model2.fit(X, y, validation_data=(Xval, yval), epochs=EPOCHS, batch_size=20)
    model2.save("fit2.tf")

    same, msg = safekeras.check_checkpoint_equality("fit.tf", "fit2.tf")
    print(msg)
    assert same is False, msg

    # coping with trashed files
    cleanup_file("fit.tf/saved_model.pb")
    same, msg = safekeras.check_checkpoint_equality("fit.tf", "fit2.tf")
    assert same is False, msg
    same, msg = safekeras.check_checkpoint_equality("fit2.tf", "fit.tf")
    assert same is False, msg

    for name in ("fit.tf", "fit2.tf", "refit.tf"):
        cleanup_file(name)


def test_load():
    """Tests the oading functionality"""

    # make a model, train then save it
    model, X, y, Xval, yval = make_small_model()
    loss = tf.keras.losses.CategoricalCrossentropy(
        from_logits=False, reduction=tf.losses.Reduction.NONE
    )
    model.compile(loss=loss, optimizer=None)
    model.fit(X, y, validation_data=(Xval, yval), epochs=EPOCHS, batch_size=20)
    model.save("keras_save.tf")

    # won't load with invalid names
    ok, _ = safekeras.load_safe_keras_model()
    assert ok is False, "can't load with no model file name"

    ok, _ = safekeras.load_safe_keras_model("keras_save.h5")
    assert ok is False, "can only load from .tf file"

    # should load fine with right name
    ok, reloaded_model = safekeras.load_safe_keras_model("keras_save.tf")
    assert ok is True
    ypred = "over-write-me"
    ypred = reloaded_model.predict(X)
    assert isinstance(ypred, np.ndarray)

    cleanup_file("keras_save.tf")
    cleanup_file("tfsaves")


def test_keras_model_created():
    "Test keras model"
    model, _, _, _, _ = make_small_model()
    rightname = "KerasModel"
    assert (
        model.model_type == rightname
    ), "failed check for model type being set in init()"
    # noise multiplier should have been reset from default to one that matches rules.json
    assert model.noise_multiplier == 0.7


def test_second_keras_model_created():
    "Test second keras model"
    X, _, _, _ = get_data()
    tf.random.set_seed(12345)
    initializer = tf.keras.initializers.Zeros()
    input_data = Input(shape=X[0].shape)
    xx = Dense(128, activation="relu", kernel_initializer=initializer)(input_data)
    xx = Dense(128, activation="relu", kernel_initializer=initializer)(xx)
    xx = Dense(64, activation="relu", kernel_initializer=initializer)(xx)
    output = Dense(n_classes, activation="softmax", kernel_initializer=initializer)(xx)
    _ = SafeKerasModel(
        inputs=input_data,
        outputs=output,
        name="test",
        num_samples=X.shape[0],
        epochs=EPOCHS,
    )
    model2 = SafeKerasModel(
        inputs=input_data,
        outputs=output,
        name="test",
        num_samples=X.shape[0],
        epochs=EPOCHS,
    )
    rightname = "KerasModel"
    assert (
        model2.model_type == rightname
    ), "failed check for second model type being set in init()"
    # noise multiplier should have been reset from default to one that matches rules.json
    assert model2.noise_multiplier == 0.7


def test_keras_model_compiled_as_DP():
    "Test Compile DP"
    model, X, _, _, _ = make_small_model()
    loss = tf.keras.losses.CategoricalCrossentropy(
        from_logits=False, reduction=tf.losses.Reduction.NONE
    )
    model.compile(loss=loss, optimizer=None)
    isDP, _ = safekeras.check_optimizer_is_DP(model.optimizer)
    assert isDP, "failed check that optimizer is dP"

    right_epsilon = 20.363059561511612
    model.check_epsilon(X.shape[0], 20, 10)
    assert (
        model.current_epsilon == right_epsilon
    ), "failed check for epsilon after compilation"

    # check this works
    ok, msg = model.check_epsilon(X.shape[0], 0, 10)
    assert ok, "should be ok after resetting batch size =1"
    correct_msg = get_reporting_string(name="division_by_zero")
    assert msg == correct_msg, msg


def test_keras_basic_fit():
    """SafeKeras using recommended values."""
    model, X, y, Xval, yval = make_small_model()

    loss = tf.keras.losses.CategoricalCrossentropy(
        from_logits=False, reduction=tf.losses.Reduction.NONE
    )
    model.compile(loss=loss, optimizer=None)

    isDP, msg = safekeras.check_optimizer_is_DP(model.optimizer)
    assert isDP, "failed check that optimizer is dP"

    # first check that stops when not DP  if they say refine
    ok, msg = model.fit(
        X,
        y,
        validation_data=(Xval, yval),
        epochs=10,
        batch_size=X.shape[0],
        refine_epsilon=True,
    )
    assert ok is False

    # now default (False)
    model.fit(X, y, validation_data=(Xval, yval), epochs=EPOCHS, batch_size=20)

    DPused, msg = safekeras.check_DP_used(model.optimizer)
    assert (
        DPused
    ), "Failed check that DP version of optimiser was actually used in training"

    loss, acc = model.evaluate(X, y)
    expected_accuracy = UNSAFE_ACC
    assert round(acc, 6) == round(
        expected_accuracy, 6
    ), "failed check that accuracy is as expected"

    msg, disclosive = model.preliminary_check()
    correct_msg = "Model parameters are within recommended ranges.\n"
    assert msg == correct_msg, "failed check params are within range"
    assert disclosive is False, "failed check disclosive is false"


def test_keras_save_actions():
    "Test save action"
    # create, compile and train model
    model, X, y, Xval, yval = make_small_model()

    loss = tf.keras.losses.CategoricalCrossentropy(
        from_logits=False, reduction=tf.losses.Reduction.NONE
    )
    model.compile(loss=loss, optimizer=None)
    model.fit(X, y, validation_data=(Xval, yval), epochs=EPOCHS, batch_size=20)

    # start with .tf and .h5 which should work
    names = ("safekeras.tf", "safekeras.h5")
    for name in names:
        # clear existing files
        cleanup_file(name)
        # save file
        model.save(name)
        assert os.path.exists(name), f"Failed test to save model as {name}"
        # clean up
        cleanup_file(name)

    # now other versions which should not
    names = ("safekeras.sav", "safekeras.pkl", "randomfilename", "undefined")
    for name in names:
        cleanup_file(name)
        model.save(name)
        assert os.path.exists(name) is False, f"Failed test NOT to save model as {name}"
        cleanup_file(name)
    # cleeanup
    cleanup_file("tfsaves")


def test_keras_unsafe_l2_norm():
    """SafeKeras using unsafe values."""
    model, X, y, Xval, yval = make_small_model()

    loss = tf.keras.losses.CategoricalCrossentropy(
        from_logits=False, reduction=tf.losses.Reduction.NONE
    )
    model.compile(loss=loss, optimizer=None)

    isDP, msg = safekeras.check_optimizer_is_DP(model.optimizer)
    assert isDP, "failed check that optimizer is dP"

    model.l2_norm_clip = 0.9

    model.fit(X, y, validation_data=(Xval, yval), epochs=EPOCHS, batch_size=20)

    DPused, msg = safekeras.check_DP_used(model.optimizer)
    assert (
        DPused
    ), "Failed check that DP version of optimiser was actually used in training"

    loss, acc = model.evaluate(X, y)
    expected_accuracy = UNSAFE_ACC
    assert round(acc, 6) == round(
        expected_accuracy, 6
    ), "failed check that accuracy is as expected"

    msg, disclosive = model.preliminary_check()
    correct_msg = (
        "WARNING: model parameters may present a disclosure risk:"
        "\n- parameter l2_norm_clip = 0.9 identified as less than the recommended "
        "min value of 1.0."
    )
    assert msg == correct_msg, "failed check correct warning message"
    assert disclosive is True, "failed check disclosive is True"


def test_keras_unsafe_noise_multiplier():
    """SafeKeras using unsafe values."""
    model, X, y, Xval, yval = make_small_model()

    loss = tf.keras.losses.CategoricalCrossentropy(
        from_logits=False, reduction=tf.losses.Reduction.NONE
    )
    model.compile(loss=loss, optimizer=None)

    isDP, msg = safekeras.check_optimizer_is_DP(model.optimizer)
    assert isDP, "failed check that optimizer is dP"

    model.noise_multiplier = 1.0

    model.fit(X, y, validation_data=(Xval, yval), epochs=EPOCHS, batch_size=20)

    DPused, msg = safekeras.check_DP_used(model.optimizer)
    assert (
        DPused
    ), "Failed check that DP version of optimiser was actually used in training"

    loss, acc = model.evaluate(X, y)
    expected_accuracy = UNSAFE_ACC
    assert round(acc, 6) == round(
        expected_accuracy, 6
    ), "failed check that accuracy is as expected"

    msg, disclosive = model.preliminary_check()
    correct_msg = (
        "WARNING: model parameters may present a disclosure risk:"
        "\n- parameter noise_multiplier = 1.0 identified as greater than the "
        "recommended max value of 0.9."
    )

    assert msg == correct_msg, "failed check params are within range"
    assert disclosive is True, "failed check disclosive is True"


def test_keras_unsafe_min_epsilon():
    """SafeKeras using unsafe values."""
    model, X, y, Xval, yval = make_small_model()

    loss = tf.keras.losses.CategoricalCrossentropy(
        from_logits=False, reduction=tf.losses.Reduction.NONE
    )
    model.compile(loss=loss, optimizer=None)

    isDP, msg = safekeras.check_optimizer_is_DP(model.optimizer)
    assert isDP, "failed check that optimizer is dP"

    model.min_epsilon = 4

    model.fit(X, y, validation_data=(Xval, yval), epochs=EPOCHS, batch_size=20)

    DPused, msg = safekeras.check_DP_used(model.optimizer)
    assert (
        DPused
    ), "Failed check that DP version of optimiser was actually used in training"

    loss, acc = model.evaluate(X, y)
    expected_accuracy = UNSAFE_ACC
    assert round(acc, 6) == round(
        expected_accuracy, 6
    ), "failed check that accuracy is as expected"

    msg, disclosive = model.preliminary_check()
    correct_msg = (
        "WARNING: model parameters may present a disclosure risk:"
        "\n- parameter min_epsilon = 4 identified as less than the recommended min value of 5."
    )

    assert msg == correct_msg, "failed check correct warning message"
    assert disclosive is True, "failed check disclosive is True"


def test_keras_unsafe_delta():
    """SafeKeras using unsafe values."""
    model, X, y, Xval, yval = make_small_model()

    loss = tf.keras.losses.CategoricalCrossentropy(
        from_logits=False, reduction=tf.losses.Reduction.NONE
    )
    model.compile(loss=loss, optimizer=None)

    isDP, msg = safekeras.check_optimizer_is_DP(model.optimizer)
    assert isDP, "failed check that optimizer is dP"

    model.delta = 1e-6

    model.fit(X, y, validation_data=(Xval, yval), epochs=EPOCHS, batch_size=20)

    DPused, msg = safekeras.check_DP_used(model.optimizer)
    assert (
        DPused
    ), "Failed check that DP version of optimiser was actually used in training"

    loss, acc = model.evaluate(X, y)
    expected_accuracy = UNSAFE_ACC
    assert round(acc, 6) == round(
        expected_accuracy, 6
    ), "failed check that accuracy is as expected"

    msg, disclosive = model.preliminary_check()
    correct_msg = (
        "WARNING: model parameters may present a disclosure risk:\n"
        "- parameter delta = 1e-06 identified as less than the recommended min value of 1e-05."
    )
    assert msg == correct_msg, "failed check params are within range"
    assert disclosive is True, "failed check disclosive is True"


def test_keras_unsafe_batch_size():
    """SafeKeras using unsafe values."""
    model, X, y, Xval, yval = make_small_model()

    loss = tf.keras.losses.CategoricalCrossentropy(
        from_logits=False, reduction=tf.losses.Reduction.NONE
    )
    model.compile(loss=loss, optimizer=None)

    isDP, msg = safekeras.check_optimizer_is_DP(model.optimizer)
    assert isDP, "failed check that optimizer is dP"

    model.batch_size = 34

    model.fit(X, y, validation_data=(Xval, yval), epochs=EPOCHS, batch_size=20)

    DPused, msg = safekeras.check_DP_used(model.optimizer)
    assert (
        DPused
    ), "Failed check that DP version of optimiser was actually used in training"

    loss, acc = model.evaluate(X, y)
    expected_accuracy = UNSAFE_ACC
    assert round(acc, 6) == round(
        expected_accuracy, 6
    ), "failed check that accuracy is as expected"

    msg, disclosive = model.preliminary_check()
    correct_msg = "Model parameters are within recommended ranges.\n"
    assert msg == correct_msg, "failed check params are within range"
    assert disclosive is False, "failed check disclosive is false"


def test_keras_unsafe_learning_rate():
    """SafeKeras using unsafe values."""
    model, X, y, Xval, yval = make_small_model()

    loss = tf.keras.losses.CategoricalCrossentropy(
        from_logits=False, reduction=tf.losses.Reduction.NONE
    )
    model.compile(loss=loss, optimizer=None)

    isDP, msg = safekeras.check_optimizer_is_DP(model.optimizer)
    assert isDP, "failed check that optimizer is dP"

    model.learning_rate = 0.2

    model.fit(X, y, validation_data=(Xval, yval), epochs=EPOCHS, batch_size=20)

    DPused, msg = safekeras.check_DP_used(model.optimizer)
    assert (
        DPused
    ), "Failed check that DP version of optimiser was actually used in training"

    loss, acc = model.evaluate(X, y)
    expected_accuracy = UNSAFE_ACC
    assert round(acc, 6) == round(
        expected_accuracy, 6
    ), "failed check that accuracy is as expected"

    msg, disclosive = model.preliminary_check()
    correct_msg = "Model parameters are within recommended ranges.\n"

    assert msg == correct_msg, "failed check warning message incorrect"
    assert disclosive is False, "failed check disclosive is false"


def test_create_checkfile():
    """Test create checkfile"""
    # create, compile and train model
    model, X, y, Xval, yval = make_small_model()

    loss = tf.keras.losses.CategoricalCrossentropy(
        from_logits=False, reduction=tf.losses.Reduction.NONE
    )
    model.compile(loss=loss, optimizer=None)
    model.fit(X, y, validation_data=(Xval, yval), epochs=EPOCHS, batch_size=20)

    # start with .tf and .h5 which should work
    names = ("safekeras.tf", "safekeras.h5")
    for name in names:
        # clear existing files
        cleanup_file(name)
        # save file
        model.save(name)
        assert os.path.exists(name), f"Failed test to save model as {name}"

        model.request_release(name)
        assert os.path.exists(name), f"Failed test to save model as {name}"

        researcher = getpass.getuser()
        outputfilename = researcher + "_checkfile.json"
        assert os.path.exists(
            outputfilename
        ), f"Failed test to save checkfile as {outputfilename}"

        # Using readlines()
        with open(outputfilename, encoding="utf-8") as file1:
            lines = file1.readlines()

        count = 0
        # Strips the newline character
        for line in lines:
            count += 1
            print(f"Line{count}: {line.strip()}")

        # clean up
        cleanup_file(name)

    # now other versions which should not
    names = ("safekeras.sav", "safekeras.pkl", "randomfilename", "undefined")
    for name in names:
        cleanup_file(name)
        model.save(name)
        assert os.path.exists(name) is False, f"Failed test NOT to save model as {name}"
        cleanup_file(name)


def test_posthoc_check():
    """testing the posthoc checking function"""
    # make a model, train then save it
    model, X, y, Xval, yval = make_small_model()
    loss = tf.keras.losses.CategoricalCrossentropy(
        from_logits=False, reduction=tf.losses.Reduction.NONE
    )
    model.compile(loss=loss, optimizer=None)
    model.fit(X, y, validation_data=(Xval, yval), epochs=1, batch_size=20)

    # should be ok
    _, disclosive = model.posthoc_check()
    assert disclosive is False, "base config in tests should be ok"

    # change optimizer and some other settings
    # in way that stresses lots of routes
    cleanup_file("tfsaves/fit_model.tf")
    model.epochs = 1000
    model.optimizer = tf.keras.optimizers.get("SGD")
    _, disclosive = model.posthoc_check()
    assert disclosive is True, "should pick up optimizer changed"

    cleanup_file("keras_save.tf")
    cleanup_file("tfsaves")


def test_final_cleanup():
    """clean up any files let around by other tests"""
    cleanup_file("tfsaves")



"""test_lira_attack.py
Copyright (C) Jim Smith2022  <james.smith@uwe.ac.uk>
"""
# pylint: disable = duplicate-code

import os
import sys

# import json
from unittest.mock import patch

# import numpy as np
from sklearn.datasets import load_breast_cancer
from sklearn.ensemble import RandomForestClassifier
from sklearn.model_selection import train_test_split

from aisdc.attacks import likelihood_attack
from aisdc.attacks.dataset import Data  # pylint: disable = import-error
from aisdc.attacks.likelihood_attack import (  # pylint: disable = import-error
    LIRAAttack,
    LIRAAttackArgs,
)


def clean_up(name):
    """removes unwanted files or directory"""
    if os.path.exists(name) and os.path.isfile(name):
        os.remove(name)


def test_lira_attack():
    """tests the lira code two ways"""
    args = LIRAAttackArgs(n_shadow_models=50, report_name="lira_example_report")
    attack_obj = LIRAAttack(args)
    attack_obj.setup_example_data()
    attack_obj.attack_from_config()
    attack_obj.example()

    X, y = load_breast_cancer(return_X_y=True, as_frame=False)
    train_X, test_X, train_y, test_y = train_test_split(X, y, test_size=0.3)
    dataset = Data()
    dataset.add_processed_data(train_X, train_y, test_X, test_y)

    # target_model = SVC(gamma=0.1, probability=True)
    target_model = RandomForestClassifier(
        n_estimators=100, min_samples_split=2, min_samples_leaf=1
    )
    target_model.fit(train_X, train_y)

    args2 = LIRAAttackArgs(n_shadow_models=50, report_name="lira_example2_report")
    attack_obj2 = LIRAAttack(args2)
    attack_obj2.attack(dataset, target_model)
    output2 = attack_obj2.make_report()  # also makes .pdf and .json files
    _ = output2["attack_metrics"][0]


def test_check_and_update_dataset():
    """test the code that removes items from test set with classes
    not present in training set"""
    args = LIRAAttackArgs(n_shadow_models=50, report_name="lira_example_report")
    attack_obj = LIRAAttack(args)
    attack_obj.setup_example_data()
    attack_obj.attack_from_config()
    attack_obj.example()

    X, y = load_breast_cancer(return_X_y=True, as_frame=False)
    train_X, test_X, train_y, test_y = train_test_split(X, y, test_size=0.3)
    # now make test[0] have a  class not present in training set
    test_y[0] = 5
    dataset = Data()
    dataset.add_processed_data(train_X, train_y, test_X, test_y)

    # make target model
    target_model = RandomForestClassifier(
        n_estimators=100, min_samples_split=2, min_samples_leaf=1
    )
    target_model.fit(train_X, train_y)
    attack_obj._check_and_update_dataset(  # pylint:disable=protected-access
        dataset, target_model
    )


def test_main():
    """test invocation via command line"""

    # option 1
    testargs = ["prog", "run-example"]
    with patch.object(sys, "argv", testargs):
        likelihood_attack.main()

    # option 2
    testargs = ["prog", "run-attack", "--j", "tests/lrconfig.json"]
    with patch.object(sys, "argv", testargs):
        likelihood_attack.main()

    # option 3 "setup-example-data"
    testargs = ["prog", "setup-example-data"]  # , "--j", "tests/lrconfig.json"]
    with patch.object(sys, "argv", testargs):
        likelihood_attack.main()


def test_cleanup():
    """cleans up various files made during the tests"""
    names = [
        "lr_report.pdf",
        "log_roc.png",
        "lr_report.json",
        "lira_example2_report.json",
        "lira_example2_report.pdf",
        "test_preds.csv",
        "config.json",
        "train_preds.csv",
        "test_data.csv",
        "train_data.csv",
    ]
    for name in names:
        clean_up(name)
